# aiozabbix - Asynchronous Zabbix API Python Library
#
# Original Ruby Library is Copyright (C) 2009 Andrew Nelson nelsonab(at)red-tux(dot)net
# Original Python Library is Copyright (C) 2009 Brett Lentz brett.lentz(at)gmail(dot)com
#
# Copyright (C) 2011-2018 Luke Cyca me(at)lukecyca(dot)com
# Copyright (C) 2018-2019 Modio AB
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import aiohttp


class ZabbixAPIException(Exception):
    pass


class ZabbixAPI:
    DEFAULT_HEADERS = {'Content-Type': 'application/json-rpc'}

    LOGIN_METHODS = ('user.login', 'user.authenticate')
    UNAUTHENTICATED_METHODS = ('apiinfo.version',) + LOGIN_METHODS
    AUTH_ERROR_FRAGMENTS = (
        'authori',                   # From CLocalApiClient::authenticate
        'permission',                # From CUser::checkAuthentication
        're-login',                  # From many places
    )

    def __init__(self,
                 server='http://localhost/zabbix',
                 *,
                 timeout=None,
                 client_session=None,
                 headers=None):

        self.url = server + '/api_jsonrpc.php'

        if client_session is None:
            self.client_session = aiohttp.ClientSession()
        else:
            self.client_session = client_session

        self.timeout = timeout

        self.auth = ''
        self.shared_state = {'next_jsonrpc_id': 0}
        self.do_login = None

        self.headers = self.DEFAULT_HEADERS.copy()
        if headers is not None:
            self.headers.update(headers)

    def with_headers(self, headers):
        """Make a copy of the ZabbixAPI object which sets extra HTTP headers.

        """
        result = ZabbixAPI.__new__(ZabbixAPI)
        result.url = self.url
        result.client_session = self.client_session
        result.timeout = self.timeout
        result.auth = self.auth
        result.shared_state = self.shared_state
        result.do_login = self.do_login
        result.headers = self.headers.copy()
        result.headers.update(headers)
        return result

    async def do_request(self, method, params=None, auth_retries=1):
        request_json = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params or {},
            'id': self.shared_state['next_jsonrpc_id'],
        }
        self.shared_state['next_jsonrpc_id'] += 1

        if method in self.UNAUTHENTICATED_METHODS:
            return await self.post_request(request_json)

        request_json['auth'] = self.auth

        try:
            return await self.post_request(request_json)
        except ZabbixAPIException as exc:
            if auth_retries > 0 and self.do_login and self.is_auth_error(exc):
                await self.do_login()
                return await self.do_request(method, params, auth_retries=auth_retries - 1)
            raise

    async def post_request(self, request_json):
        response = await self.client_session.post(self.url,
                                                  json=request_json,
                                                  headers=self.headers,
                                                  timeout=self.timeout)
        response.raise_for_status()

        try:
            response_json = await response.json()
        except ValueError as exc:
            raise ZabbixAPIException(f'Unable to parse JSON: {response.text()}') from exc

        if 'error' in response_json:
            # Workaround for ZBX-9340, some errors don't contain 'data':
            if 'data' not in response_json['error']:
                response_json['error']['data'] = 'ZBX-9340: No data'

            err = response_json['error']
            msg = f"Error {err['code']}: {err['message']}, {err['data']}"
            raise ZabbixAPIException(msg, err['code'])

        return response_json

    @classmethod
    def method_needs_auth(cls, method):
        return method not in cls.UNAUTHENTICATED_METHODS

    @classmethod
    def is_auth_error(cls, exc):
        """Determine if an error is an authorization error.

        This makes a best effort attempt to recognize authentication
        or authorization errors. Unfortunately the general JSON-RPC
        error code -32602 (Invalid params) and the generic Zabbix
        error -32500 are used for these types of errors.

        The error messages may also be localized which could make this
        check fail.
        """

        err = str(exc).lower()
        return any(x in err for x in cls.AUTH_ERROR_FRAGMENTS)

    async def login(self, user='', password=''):
        async def do_login():
            self.auth = ''
            self.auth = await self.user.login(user=user, password=password)

        self.do_login = do_login
        await self.do_login()

    async def confimport(self, confformat='', source='', rules=''):
        return await self.configuration.import_(format=confformat, source=source, rules=rules)

    def __getattr__(self, attr):
        return ZabbixAPIObjectClass(name=attr, parent=self)


class ZabbixAPIObjectClass:
    def __init__(self, *, name, parent):
        self.name = name
        self.parent = parent

    def __getattr__(self, attr):
        if attr == 'import_':
            attr = 'import'

        async def method(*args, **kwargs):
            if args and kwargs:
                raise TypeError(
                    'Method may be called with positional arguments or '
                    'keyword arguments, but not both at the same time'
                )

            response = await self.parent.do_request(f'{self.name}.{attr}', args or kwargs)
            return response['result']

        return method



from aiohttp import web
import pytest

from aiozabbix import ZabbixAPI, ZabbixAPIException


async def mock_jsonrpc(request):
    request.app["requests"].append(request)

    request_json = await request.json()
    assert request_json["jsonrpc"] == "2.0"
    assert "params" in request_json
    assert "id" in request_json

    method = request_json["method"]
    if ZabbixAPI.method_needs_auth(method):
        if not request.app["logged_in"]:
            return web.json_response(
                {
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32602,
                        "message": "Invalid params.",
                        "data": "Not authorised.",
                    },
                    "id": request_json["id"],
                }
            )

    if method == "user.login":
        request.app["logged_in"] = True

    return web.json_response(
        {
            "jsonrpc": "2.0",
            "result": f"mock response for {method}",
            "id": request_json["id"],
        }
    )


@pytest.fixture
def mock_server_app():
    app = web.Application()
    app.router.add_post("/api_jsonrpc.php", mock_jsonrpc)
    app["requests"] = []
    app["logged_in"] = False
    return app


@pytest.fixture
async def client_session(aiohttp_client, mock_server_app):
    return await aiohttp_client(mock_server_app)


@pytest.fixture
async def zapi(client_session, mock_server_app):
    return ZabbixAPI(server="", client_session=client_session)


async def test_unauthenticated_calls_should_work(mock_server_app, zapi):
    await zapi.apiinfo.version()

    requests = [await r.json() for r in mock_server_app["requests"]]
    assert len(requests) == 1
    assert requests[0]["method"] == "apiinfo.version"


async def test_authenticated_mock_calls_should_fail_before_login(zapi):
    with pytest.raises(ZabbixAPIException):
        await zapi.hostgroup.get()


async def test_login_should_work(mock_server_app, zapi):
    await zapi.login(user="Admin", password="zabbix")

    requests = [await r.json() for r in mock_server_app["requests"]]
    assert len(requests) == 1
    assert requests[0]["method"] == "user.login"
    assert requests[0]["params"] == {"user": "Admin", "password": "zabbix"}


async def test_authenticated_mock_calls_should_succeed_after_login(zapi):
    await zapi.login(user="Admin", password="zabbix")
    response = await zapi.hostgroup.get()
    assert "error" not in response


async def test_auth_error_should_cause_auto_relogin(mock_server_app, zapi):
    await zapi.login(user="Admin", password="zabbix")
    await zapi.hostgroup.get()

    mock_server_app["logged_in"] = False

    await zapi.hostgroup.get()

    requests = [await r.json() for r in mock_server_app["requests"]]
    assert len(requests) == 5
    assert requests[0]["method"] == "user.login"
    assert requests[1]["method"] == "hostgroup.get"
    # Second hostgroup.get that fails due to being logged out
    assert requests[2]["method"] == "hostgroup.get"
    # Auto login
    assert requests[3]["method"] == "user.login"
    # Retry of hostgroup.get
    assert requests[4]["method"] == "hostgroup.get"


async def test_import_underscore_attr_should_be_rewritten(mock_server_app, zapi):
    await zapi.login(user="Admin", password="zabbix")
    await zapi.confimport(
        confformat="xml", rules={}, source="<zabbix_export>...</zabbix_export>"
    )

    requests = [await r.json() for r in mock_server_app["requests"]]
    assert len(requests) == 2
    assert requests[1]["method"] == "configuration.import"
    assert requests[1]["params"] == {
        "format": "xml",
        "rules": {},
        "source": "<zabbix_export>...</zabbix_export>",
    }


async def test_custom_headers_should_be_sent(mock_server_app, client_session):
    zapi = ZabbixAPI(
        server="", client_session=client_session, headers={"User-Agent": "zabbixapp"}
    )

    await zapi.apiinfo.version()

    requests = mock_server_app["requests"]
    assert len(requests) == 1
    assert requests[0].headers["User-Agent"] == "zabbixapp"


async def test_zabbix_api_copies_should_share_state_correctly(mock_server_app, client_session):
    zapi = ZabbixAPI(
        server="", client_session=client_session, headers={"User-Agent": "zabbixapp"}
    )

    await zapi.apiinfo.version()

    zapi_with_extra_header = zapi.with_headers({"X-Extra-Header": "Yes"})
    await zapi_with_extra_header.apiinfo.version()

    await zapi.apiinfo.version()

    requests = mock_server_app["requests"]
    assert len(requests) == 3

    assert requests[0].headers["User-Agent"] == "zabbixapp"
    assert "X-Extra-Header" not in requests[0].headers

    assert requests[1].headers["User-Agent"] == "zabbixapp"
    assert requests[1].headers["X-Extra-Header"] == "Yes"

    assert requests[2].headers["User-Agent"] == "zabbixapp"
    assert "X-Extra-Header" not in requests[2].headers

    request_ids = [(await r.json())["id"] for r in mock_server_app["requests"]]
    assert request_ids == [0, 1, 2]



from setuptools import setup

setup(
    name='aiozabbix',
    version='1.1.0',
    description='Asynchronous Zabbix API Python interface',
    url='http://gitlab.com/ModioAB/aiozabbix',
    author='Modio AB',
    author_email='nili@modio.se',
    license='LGPL',
    classifiers=[
        'Framework :: AsyncIO',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='zabbix monitoring api',
    packages=['aiozabbix'],
    python_requires='>=3.6',
    install_requires=[
        'aiohttp',
    ],
    extras_require={
        'testing': [
            'flake8',
            'pytest',
            'pytest-aiohttp',
            'pytest-cov',
        ]
    }
)
